/*      Copyright (c) 1995 Mix Software, Inc.

        blockio.c - module to read and write blocks from an open file
                (internal use only)

        The file is assumed to be a collection of blksize-sized blocks.
        The buffer must be large enough to hold blksize bytes.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include "cdtdef.h"
#include "blockio.h"

#define STDOUT_HANDLE 1
#define USE_STD_IO    1    /* set to 0 to remove references to fputs */

/* define 32 bit registers for int86 */
#if defined(M_I386) || defined(__386__)
   #define AX x.eax
   #define BX x.ebx
   #define CX x.ecx
   #define DX x.edx
   #define SI x.esi
   #define DI x.edi
   #define int86 int386
   #define SET_FS(s) s.fs = s.ds
   #define SET_GS(s) s.gs = s.ds
#else
   #define AX x.ax
   #define BX x.bx
   #define CX x.cx
   #define DX x.dx
   #define SI x.si
   #define DI x.di
   #define SET_FS(s)
   #define SET_GS(s)
#endif

int _cbt_doserrno = 0;
int _b_no_flush = 0;

#define UPPER_WORD(value) ((unsigned)((value) >> 16));
#define LOWER_WORD(value) ((unsigned)((value) & 0xffff));

/*
        Defined functions (CBT internal use only)

        cb_openfile               open a random file
        cb_close                  close a file
        cb_read                   read a random area from a file
        cb_readblk                read block from file
        cb_readseq                read sequentially
        cb_write                  write to a file
        cb_writeblk               write block to file
        cb_writeseq               write sequentially
        cb_commitblk              commit updates to disk
        cb_lock_area              lock a portion of a file
        cb_unlock_area            unlock a portion of a file
        cb_delay                  delay (before lock retries)
        cb_prints                 print a string to stdout
        cb_can_share              can files be shared?
*/
int cb_read(int fd, unsigned long location, int size, void *buffer);
int cb_write(int fd, unsigned long location, int size, void *buffer);

/**/
/* ------------------------------------------------------------------------
        cb_openfile - open a block oriented file

        Filename is the name of the file to be opened.  Share is either I_YES
        or I_NO, (1 or 0).  If share is I_YES, the file is opened for shared
        access so that multiple programs can access the file at the same time.

        Return a non-negative integer on success or -1 on error.  In the
        event of an error, errno is also set.
*/

int cb_openfile(char *filename, int mode, int share)
{
   union REGS r;
   struct SREGS sr;
   int modeflg;

   r.h.ah = 0x3d;
   if (mode == CBT_OPEN_RDONLY) modeflg = 0x80; else modeflg = 0x82;
   if (share) modeflg |= 0x40;  /* deny none */
   else modeflg |= 0x10;        /* deny all */
   r.h.al = (unsigned char) modeflg;
   segread(&sr);
   if (sizeof(char *) == sizeof(unsigned)) r.DX = (unsigned) filename;
   else {
      r.DX = FP_OFF(filename);
      sr.ds = FP_SEG(filename);
      }
   intdosx(&r, &r, &sr);
   if (r.x.cflag != 0) {
      _cbt_doserrno = r.h.al;
      return -1;
      }
   return r.AX;
} /* cbopenfile */
/**/
/* ------------------------------------------------------------------------
        cb_create - create a file and open it for exclusive write

        Filename is the name of the file to be opened.
*/

int cb_create(char *filename)
{
   union REGS r;
   struct SREGS sr;

   r.h.ah = 0x3c;
   r.CX = 0;
   segread(&sr);
   if (sizeof(char *) == sizeof(unsigned)) r.DX = (unsigned) filename;
   else {
      r.DX = FP_OFF(filename);
      sr.ds = FP_SEG(filename);
      }
   intdosx(&r, &r, &sr);
   if (r.x.cflag != 0) {
      _cbt_doserrno = r.h.al;
      return -1;
      }
   return r.AX;
} /* cb_create */
/**/
/* ------------------------------------------------------------------------

        cbcloseblk - close a block oriented file

        Return 0 on success or -1 on error.
*/

int cb_closeblk(int fd)
{
   union REGS r;
   r.h.ah = 0x3e;
   r.BX = fd;
   intdos(&r, &r);
   if (r.x.cflag != 0) {
      _cbt_doserrno = r.h.al;
      return -1;
      }
   return 0;
} /* cb_closeblk */
/**/
/* ------------------------------------------------------------------------

        cbreadblk - read one block of blksize from file into buffer

        Return OK or ERROR.
*/
int cb_readblk(int fd, Blk_Nbr blocknbr, int blksize, void *buffer)
{
   return cb_read(fd, blocknbr * blksize, blksize, buffer);
   }

/* ------------------------------------------------------------------------

        cb_read  - random read

        Return OK or ERROR.
*/

int cb_read(int fd, unsigned long location, int size, void *buffer)
{
   union REGS r;
   struct SREGS sr;

   r.h.ah = 0x42;                /* seek to position */
   r.h.al = 0;
   r.BX = fd;
   r.CX = UPPER_WORD(location);
   r.DX = LOWER_WORD(location);
   intdos(&r, &r);
   if (r.x.cflag != 0) {
      _cbt_doserrno = r.h.al;
      return I_ERROR;
      }

   r.h.ah = 0x3f;                /* read file */
   r.BX = fd;
   r.CX = size;
   segread(&sr);
   if (sizeof(char *) == sizeof(unsigned)) r.DX = (unsigned) buffer;
   else {
      r.DX = FP_OFF(buffer);
      sr.ds = FP_SEG(buffer);
      }
   intdosx(&r, &r, &sr);
   if (r.x.cflag != 0) {
      _cbt_doserrno = r.h.al;
      return I_ERROR;
      }
   if (r.AX < (unsigned) size) {
      _cbt_doserrno = EC_DOS_READERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cbreadblk */
/* ------------------------------------------------------------------------

        cb_readseq  - sequential read

        Return OK or ERROR.
*/

int cb_readseq(int fd, int size, void *buffer)
{
   union REGS r;
   struct SREGS sr;

   r.h.ah = 0x3f;                /* read file */
   r.BX = fd;
   r.CX = size;
   segread(&sr);
   if (sizeof(char *) == sizeof(unsigned)) r.DX = (unsigned) buffer;
   else {
      r.DX = FP_OFF(buffer);
      sr.ds = FP_SEG(buffer);
      }
   intdosx(&r, &r, &sr);
   if (r.x.cflag != 0) {
      _cbt_doserrno = r.h.al;
      return I_ERROR;
      }
   if (r.AX < (unsigned) size) {
      _cbt_doserrno = EC_DOS_READERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cbreadblk */
/**/
/* ------------------------------------------------------------------------

        cbwriteblk - write one block of blksize to file from buffer

        Return OK or ERROR.
*/

int cb_writeblk(int fd, Blk_Nbr blocknbr, int blksize, void *buffer)
{
   return cb_write(fd, blocknbr * blksize, blksize, buffer);
   }

/* ------------------------------------------------------------------------

        cb_write - random write

        Return OK or ERROR.
*/

int cb_write(int fd, unsigned long location, int size, void *buffer)
{
   union REGS r;
   struct SREGS sr;

   r.h.ah = 0x42;                /* seek to position */
   r.h.al = 0;
   r.BX = fd;
   r.CX = UPPER_WORD(location);
   r.DX = LOWER_WORD(location);
   intdos(&r, &r);
   if (r.x.cflag != 0) {
      _cbt_doserrno = r.h.al;
      return I_ERROR;
      }

   r.h.ah = 0x40;                /* write to file */
   r.BX = fd;
   r.CX = size;
   segread(&sr);
   if (sizeof(char *) == sizeof(unsigned)) r.DX = (unsigned) buffer;
   else {
      r.DX = FP_OFF(buffer);
      sr.ds = FP_SEG(buffer);
      }
   intdosx(&r, &r, &sr);
   if (r.x.cflag != 0) {
      _cbt_doserrno = r.h.al;
      return I_ERROR;
      }
   if (r.AX < (unsigned) size) {
      _cbt_doserrno = EC_DOS_WRITERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cb_write */

/* ------------------------------------------------------------------------

        cb_writeseq - sequential write

        Return OK or ERROR.
*/

int cb_writeseq(int fd, int size, void *buffer)
{
   union REGS r;
   struct SREGS sr;

   r.h.ah = 0x40;
   r.BX = fd;
   r.CX = size;
   segread(&sr);
   if (sizeof(char *) == sizeof(unsigned)) r.DX = (unsigned) buffer;
   else {
      r.DX = FP_OFF(buffer);
      sr.ds = FP_SEG(buffer);
      }
   intdosx(&r, &r, &sr);
   if (r.x.cflag != 0) {
      _cbt_doserrno = r.h.al;
      return I_ERROR;
      }
   if (r.AX < (unsigned) size) {
      _cbt_doserrno = EC_DOS_WRITERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cb_write */

/**/
/* ------------------------------------------------------------------------

        cb_remove - delete a file

        Return OK or ERROR.
*/

int cb_remove(char *filename)
{
   union REGS r;
   struct SREGS sr;

   r.h.ah = 0x41;
   segread(&sr);
   if (sizeof(char *) == sizeof(unsigned)) r.DX = (unsigned) filename;
   else {
      r.DX = FP_OFF(filename);
      sr.ds = FP_SEG(filename);
      }
   intdosx(&r, &r, &sr);
   if (r.x.cflag != 0) {
      _cbt_doserrno = r.h.al;
      return I_ERROR;
      }
   return I_OK;
   } /* cb_remove */

/**/
/* ------------------------------------------------------------------------

        cb_lock_area - put a record lock on a file region

        Return OK or ERROR.
*/

int cb_lock_area(int fd, unsigned long location, long size, int tries,
   long delay, long timeout)
{
   union REGS regs, r1, r2;
   long  stoptime;
   long  time, lasttime;
   int   hour;
   static int multitasking = 1;
   static int bios15 = 1;

   /* set registers for lock system call */
   regs.h.ah = 0x5c;
   regs.h.al = 0;
   regs.BX = fd;
   regs.CX = UPPER_WORD(location);
   regs.DX = LOWER_WORD(location);
   regs.SI = UPPER_WORD(size);
   regs.DI = LOWER_WORD(size);
   /* try to lock */
   intdos(&regs, &r2);
   if (r2.x.cflag == 0) return (I_OK);

   /* lock failed - retry based on arguments */
   if (tries < 2) return I_ERROR;
   if (multitasking != 0) {
      r1.AX = 0x1680;    /* give up time slice */
      int86(0x2f, &r1, &r1);
      if (r1.h.al != 0) multitasking = 0;  /* not supported */
      }
   r1.h.ah = 0x2c;   /* read system time */
   intdos(&r1, &r1);
   time = r1.h.ch*3600000L + r1.h.cl*60000L + r1.h.dh*1000L + r1.h.dl*10L;
   hour = r1.h.ch;
   stoptime = time + timeout;

   while (tries-- > 0) {
      lasttime = time;
      if (multitasking || ! bios15) {
         while (time < stoptime && (time-lasttime) < delay) {
            r1.h.ah = 0x2c;   /* read system time */
            intdos(&r1, &r1);
            time = r1.h.ch*3600000L + r1.h.cl*60000L + r1.h.dh*1000L + r1.h.dl*10L;
            if ((int) r1.h.ch < hour) time += 86400000L;  /* passed midnight */
            if (multitasking && (time-lasttime) < delay) { /* yield */
               r1.AX = 0x1680;    /* give up time slice */
               int86(0x2f, &r1, &r1);
               }
            }
         }
      else {   /* use bios15 time delay */
         time = delay * 1000L;
         r1.h.ah = 0x86;
         r1.CX = UPPER_WORD(time);
         r1.DX = LOWER_WORD(time);
         int86(0x15, &r1, &r1);
         if (r1.x.cflag != 0) bios15 = 0; /* not supported */
         }
      intdos(&regs, &r2);  /* retry the lock */
      if (r2.x.cflag == 0) return (I_OK);
      if (time > stoptime) break;    /* reached timeout */
      }  /* tries */

   return (I_ERROR);
} /* cb_lock_area */
/**/
/* ------------------------------------------------------------------------

        cb_unlock_area - remove a record lock from a file region

        Return OK or ERROR.
*/

int cb_unlock_area(int fd, unsigned long location, long size)
{
   union REGS regs;

   regs.h.ah = 0x5c;
   regs.h.al = 1;
   regs.BX = fd;
   regs.CX = UPPER_WORD(location);
   regs.DX = LOWER_WORD(location);
   regs.SI = UPPER_WORD(size);
   regs.DI = LOWER_WORD(size);
   intdos(&regs, &regs);
   if (regs.x.cflag != 0) return I_ERROR;
   return (I_OK);
} /* cb_unlock_area */
/**/
/* ------------------------------------------------------------------------

        cb_commitblk -   flush any system buffer associated with specified
                         file descriptor

        This subroutine exists to get around a serious DOS design flaw.  When
        a disk file is flushed, the data is written to the disk but neither
        the file control block (FCB) nor the file allocation table (FAT) are
        updated.  The data is on the disk but the file system doesn't know
        it is there.  In DOS 3, they added a "commit file" DOS function to
        get around this problem.  Unfortunately, not everyone is running DOS
        3 or later.  If a version less than 3.3 is found, this routine
        duplicates the file handle and close the duplicate.  This is somewhat
        slower than using the commit file function but has the benefit of
        working on all versions of DOS.

        Return OK or ERROR.
*/

int cb_commitblk(int fd)
{
   union REGS regs;
   if (_b_no_flush) return I_OK;
   if (_osmajor >= 3) {
      regs.h.ah = 0x68;
      regs.BX = fd;
      intdos(&regs, &regs);  /* flush system buffer */
      if (regs.x.cflag != 0 && regs.h.ah != 0x68) return I_ERROR;
      }
   else {
      regs.h.ah = 0x45;
      regs.BX = fd;
      intdos(&regs, &regs);
      if (regs.x.cflag != 0) {
         _cbt_doserrno = regs.AX;
         return (I_ERROR);
         }
      regs.BX = regs.AX;      /* new file handle */
      regs.h.ah = 0x3e;
      intdos(&regs, &regs);
      if (regs.x.cflag != 0) {
         _cbt_doserrno = regs.AX;
         return (I_ERROR);
         }
      }
   return I_OK;
   }
/**/
/* ------------------------------------------------------------------------

        cb_delay - time delay before lock retry

        always returns OK
*/

int cb_delay(long time)
{
   union REGS regs, r;
   long  microtime;
   long  stoptime;
   int   hour;
   static int multitasking = 1;
   static int bios15 = 1;

   if (multitasking != 0) {
      regs.AX = 0x1680;    /* give up time slice */
      int86(0x2f, &regs, &regs);
      if (regs.h.al != 0) multitasking = 0;  /* not supported */
      }
   if (bios15) {
      microtime = time * 1000L;
      regs.h.ah = 0x86;
      regs.CX = UPPER_WORD(microtime);
      regs.DX = LOWER_WORD(microtime);
      int86(0x15, &regs, &regs);
      if (regs.x.cflag != 0) bios15 = 0; /* not supported */
      }
   if (bios15 == 0) {  /* bios 15 delay not supported */
      regs.h.ah = 0x2c;
      intdos(&regs, &regs);
      stoptime = regs.h.ch*3600000L + regs.h.cl*60000L
                 + regs.h.dh*1000L + regs.h.dl*10L + time;
      hour = regs.h.ch;
      do {
         if (multitasking != 0) {
            regs.AX = 0x1680;    /* give up time slice */
            int86(0x2f, &regs, &regs);
            }
         r.h.ah = 0x2c;
         intdos(&r, &r);
         microtime = r.h.ch*3600000L + r.h.cl*60000L
                     + r.h.dh*1000L + r.h.dl*10L;
         if ( (int)r.h.ch < hour) time += 86400000L;  /* passed midnight */
         } while (stoptime > microtime);
      }
   return (I_OK);
} /* cb_delay */

/**/
/* ------------------------------------------------------------------------

        cb_prints - print a string to stdout

*/
#if USE_STD_IO
int I_ENTRY cb_prints(const char *string)
{
    if (fputs(string, stdout) != 0)
        return I_ERROR;

   return I_OK;
} /* cb_prints */

#else
int I_ENTRY cb_prints(const char *string)
{
   union REGS r, r2;
   struct SREGS sr;
   char *ptr;
   char c, lastc;

   if (string == NULL) return I_ERROR;
   segread(&sr);
   r.h.ah = 0x40;
   r.BX = STDOUT_HANDLE;
   if (sizeof(char *) == sizeof(unsigned)) r.DX = (unsigned) string;
   else {
      r.DX = FP_OFF(string);
      sr.ds = FP_SEG(string);
      }

   ptr = strchr(string, '\n');
   if (ptr == NULL) {      /* no newlines to translate */
      r.CX = strlen(string);
      intdosx(&r, &r2, &sr);
      if (r2.x.cflag != 0) {
         _cbt_doserrno = r2.AX;
         return (I_ERROR);
         }
      }
   else {
      if (ptr != string) {
         r.CX = ptr - string;
         intdosx(&r, &r2, &sr);
         if (r2.x.cflag != 0) { _cbt_doserrno = r2.AX; return (I_ERROR); }
         lastc = *(ptr-1);
         }
      else lastc = 0;
      r.CX = 1;
      r.DX = (unsigned) &c;
      while (*ptr != '\0') {
         c = *ptr++;
         if (c == '\n' && lastc != '\r') {
            lastc = '\r';
            r.DX = (unsigned) &lastc;
            intdos(&r, &r2);
            if (r2.x.cflag != 0) { _cbt_doserrno = r2.AX; return (I_ERROR); }
            r.DX = (unsigned) &c;
            }
         intdos(&r, &r2);
         if (r2.x.cflag != 0) { _cbt_doserrno = r2.AX; return (I_ERROR); }
         lastc = c;
         }
      }
   return I_OK;
} /* cb_prints */
#endif
/**/

/* ------------------------------------------------------------------------

        cb_can_share - can files be shared?

        Return I_YES or I_NO.
*/

int cb_can_share(void)
{
   union REGS r;
   static int share = -1;
   if (share == -1) {
      r.AX = 0x1000;
      int86(0x2f, &r, &r);
      if (r.h.al != 0xff) return 0;
      /* when running under MS Windows with share not loaded, windows  */
      /* reports that share is loaded, but file sharing does not work. */
      /* Test by attempting to unlock a block in stdin.  If share is   */
      /* not functioning, an error code of 1 (invalid function number) */
      /* is returned.  If share is present, it returns error code 0x21 */
      /* (no such lock) or 0x06 (invalid handle). */
      r.AX = 0x5c01;
      r.BX = 1;
      r.CX = r.DX = r.SI = r.DI = 0;
      int86(0x21, &r, &r);
      share = (r.h.al == 1) ? I_NO : I_YES;
      }
   return share;
} /* cb_can_share */
/**/

/* ------------------------------------------------------------------------

        cb_filesize - determine the size of an open file

        Return OK or ERROR.
*/

int cb_filesize(int fd, long *size)
{
   long pos;                    /* position of block */
   union REGS r;

   r.h.ah = 0x42;                /* seek to position */
   r.h.al = 0x02;
   r.BX = fd;
   r.CX = 0;
   r.DX = 0;
   intdos(&r, &r);
   if (r.x.cflag != 0) {
      _cbt_doserrno = r.h.al;
      return I_ERROR;
      }
   if (sizeof(unsigned) > 2) pos = ((r.DX & 0xffff) << 16) | (r.AX & 0xffff);
   else pos = ((long)r.DX << 16) | r.AX;
   *size = pos;
   return (I_OK);
   } /* cb_filesize */

/* ------------------------------------------------------------------------

        cbset_print - set the print function
        Return OK or ERROR.
*/

I_EXPORT Puts_Fn_Ptr I_ENTRY cbset_print(Puts_Fn_Ptr printfn)
{
   Puts_Fn_Ptr old;
   old = _cb_prints;
   _cb_prints = printfn;
   return old;
   } /* cbset_print */

/* end blockio.c */
