/*      Copyright (c) 1987 Shrier and Deihl. Licensed to Mix Software, Inc.
        Copyright (c) 1995 Mix Software, Inc. Multi-user version

        cbinsert.c - CBT cbinsert and supporting functions
*/
#include "btree.h"

/*
        Defined Functions (INTERNAL functions for use by CBT only)

        cbinsert                insert key/item into tree
        B_ins_new_tree          INTERNAL insert key into empty tree
        B_ins_tree              INTERNAL insert key into tree
        ins_subtree             INTERNAL insert key into subtree
        add_children            INTERNAL add children to node
        loc_start               INTERNAL locate start of re-sync
        loc_stop                INTERNAL locate end of re-sync
        B_add_low_key           INTERNAL add key from low path leaf
*/
static int ins_subtree(Cbtree *tree, Key_Info *key_info, Blk_Nbr blk_nbr,
                       Blk_Nbr parent, int par_pos);
static int add_children(Cbtree *tree, Blk_Nbr blk_nbr, int pos,
                        Blk_Nbr parent, int par_pos);
static int loc_start(Cbtree *tree, Node *par_node, int par_pos, int *start,
                     Blk_Nbr *child);
static int loc_stop(Cbtree *tree, Node *par_node, int start,
                    Blk_Nbr *sync_child);

/*      cbinsert - insert key and item into a tree
        If dup is I_YES, duplicate keys are allowed.
        Return OK or ERROR.
*/

I_EXPORT int I_ENTRY cbinsert(Cbtree *tree, char *key, int keylen,
      Item item, int dup)
{
   int status;                /* subroutine return status */
   Key_Info key_info;         /* key information structure */
   Ctlrec *ctl;               /* control record */

   SET_FC(FC_CBINSERT)
   if ((B_is_tree_valid(tree) == I_NO) || (B_is_ctl_valid(tree->ctl) == I_NO)
      || (B_is_key_valid(keylen) == I_NO)) return (I_ERROR);
   if (tree->readonly) ERR1X(EC_READONLY)
   key_info.key = key;
   key_info.keylen = (Keylen) keylen;
   key_info.item = item;
   key_info.dup = (short) dup;
   ctl = tree->ctl;

   /* write lock the tree for the duration of the operation */
   if (B_write_lock(ctl) != I_OK) return (I_ERROR);

   /* synchronize with the file and flush buffers if necessary */
   if (B_ctl_sync(ctl) == I_ERROR) {
      B_ctl_unlock(ctl);
      return (I_ERROR);
      }
   /*
           there are 2 cases for insert, inserting into an empty
           tree or inserting into a tree that already contains
           some key/item pairs.
   */
   if ((tree->ctl)->disk.root == (Blk_Nbr) 0)
      status = B_ins_new_tree(tree, &key_info);
   else status = B_ins_tree(tree, &key_info);
   ctl->disk.keycount++;
   B_ctl_resync(ctl, I_YES, I_YES); /* write buffers & ctl record to file */
   return (status);
   } /* cbinsert */
/**/
/*
        B_ins_new_tree - insert key and item into new, empty tree
        Return OK or ERROR.
*/

int B_ins_new_tree(Cbtree *tree, Key_Info *key_info)
{
   Node *first_leaf;       /* node buffer of first leaf node */
   Blk_Nbr root;           /* block number of root node */
   Ctlrec *ctl;            /* control structure for tree */

   ctl = tree->ctl;
   if ((first_leaf = (Node *) B_get_empty_block(tree, &root)) == (Node *) NULL) {
      SET_CC(CC_FIRSTLEAF)
      return (I_ERROR);
      }
   B_init_leaf(first_leaf, (Blk_Nbr) 0, (Blk_Nbr) 0, ctl->disk.blksize);
   if (put_blk(ctl->bfile, (char *) first_leaf, I_NO) != I_OK)
      ERR2X(EC_PUTBLK, CC_FIRSTLEAF)
   ctl->disk.root = root;
   ctl->disk.lowleaf = root;
   ctl->disk.highleaf = root;
   ctl->disk.depth = 1;
   if (B_ins_node(tree, key_info, root, (Blk_Nbr) 0, 0) != I_OK)
      return (I_ERROR);
   return (I_OK);
   } /* B_ins_new_tree */
/**/
/*
        B_ins_tree - insert key and item into existing tree

        Return OK or ERROR.
*/


int B_ins_tree(Cbtree *tree, Key_Info *key_info)
   /* tree     = tree to insert key and item into */
   /* key_info = key and item to insert */
{
   Ctlrec *ctl;            /* control record for tree */
   Blk_Nbr low_path;       /* low path of root node */
   int status;             /* subroutine return status */

   ctl = tree->ctl;
   if (B_get_low_path(tree, ctl->disk.root, &low_path) == I_ERROR)
      return (I_ERROR);
   if (low_path == LEAF_NODE) {  /* a tree with only one node */
      if ((status = B_ins_node(tree, key_info, ctl->disk.root, (Blk_Nbr) 0, 0))
         == I_ERROR) return (I_ERROR);
      }
   else if (low_path == UNUSED_NODE) ERR2X(EC_CORRUPT, CC_INSNODE)
   else {   /* start recursion */
      if ((status = ins_subtree(tree, key_info, ctl->disk.root, (Blk_Nbr) 0, 0))
         == I_ERROR) return (I_ERROR);
      }
   switch (status) {
      case SPLIT:
      case SHIFT:
      case SHIFT_INS:
         if (B_make_root(tree) == I_ERROR) return (I_ERROR);
         break;
      }
   return (I_OK);
   } /* B_ins_tree */
/**/
/*
        ins_subtree - insert key in subtree, propagate node splits

        Return OK, SPLIT, or ERROR.
*/


static int ins_subtree(Cbtree *tree, Key_Info *key_info, Blk_Nbr blk_nbr,
                       Blk_Nbr parent, int par_pos)
   /* tree     = tree containing subtree */
   /* key_info = key to insert */
   /* blk_nbr  = block number of target subtree */
   /* parent   = block number of parent node */
   /* par_pos  = position of target in parent */
{
   int status;                /* insertion status */
   int pos;                   /* key position index */
   Blk_Nbr child;             /* block number of child */
   Blk_Nbr low_path;          /* low path in child */

   /* find the insertion position at this level */
   if ((child = B_loc_subtree(tree, key_info, blk_nbr, &pos))
      == (Blk_Nbr) I_ERROR) return (I_ERROR);
   if (B_get_low_path(tree, child, &low_path) == I_ERROR)
      return (I_ERROR);
   if (low_path == LEAF_NODE) {  /* insert here */
      if ((status = B_ins_node(tree, key_info, child, blk_nbr, pos))
         == I_ERROR) return (I_ERROR);
      }
   else if (low_path == UNUSED_NODE) ERR2X(EC_CORRUPT, CC_INSNODE)
   else {   /* move to next level */
      if ((status = ins_subtree(tree, key_info, child, blk_nbr, pos))
         == I_ERROR) return (I_ERROR);
      }
   switch (status) {
      case SPLIT:
      case SHIFT:
      case SHIFT_INS:
         status = add_children(tree, blk_nbr, pos, parent, par_pos);
         break;
      }
   return (status);
   } /* ins_subtree */
/**/
/*
        add_children - add children to node

        Add_children updates the keys in a node to point to its children
        after a node split or key shift.  Add_children starts with the
        node that is at pos - 2 and scans the child nodes using the right
        link until it is sure all the children are linked into the parent.

        Return SPLIT, OK, or ERROR.
*/

static int add_children(Cbtree *tree, Blk_Nbr blk_nbr, int pos,
                        Blk_Nbr parent, int par_pos)
   /* tree    = tree containing node to add to */
   /* blk_nbr = block to add keys to */
   /* pos     = position of child to add */
   /* parent  = parent of block to add keys to */
   /* par_pos = position in parent of blk_nbr */
{
   Node *node;             /* target node buffer */
   int start;              /* start position for add */
   Blk_Nbr child;          /* block number of child */
   Blk_Nbr sync_child;     /* number of sync child */
   int status;             /* subroutine return status */
   int r_status;           /* add status */
   Bfile *bfile;           /* buffered file pointer */

   bfile = (tree->ctl)->bfile;
   r_status = I_OK;
   if ((node = (Node *) get_blk(bfile, blk_nbr)) == (Node *) NULL)
      ERR2X(EC_GETBLK, CC_RESYNCPARENT)
   if (loc_start(tree, node, pos, &start, &child) == I_ERROR) {
      rlse_blk(bfile, (char *) node);
      return (I_ERROR);
      }
   if (loc_stop(tree, node, start, &sync_child) == I_ERROR) {
      rlse_blk(bfile, (char *) node);
      return (I_ERROR);
      }
   while (child != sync_child) {
      CB_ASSERT(child != (Blk_Nbr) 0);
      if ((status =
           B_add_low_key(tree, blk_nbr, node, child, parent, par_pos, &pos))
         == I_ERROR) {
         rlse_blk(bfile, (char *) node);
         return (I_ERROR);
         }
      switch (status) {
         case SPLIT:
         case SHIFT:
         case SHIFT_INS:
            if (pos >= node->key_cnt) {
               blk_nbr = node->right;
               if (rlse_blk(bfile, (char *) node) == I_ERROR)
                  ERR2X(EC_RLSEBLK, CC_RESYNCPARENT)
               if ((node = (Node *) get_blk(bfile, blk_nbr)) == (Node *) NULL)
                  ERR2X(EC_GETBLK, CC_RESYNCPARENT)
               }
            r_status = status;
            break;
         }
      if (B_get_right(tree, child, &child) == I_ERROR) {
         rlse_blk(bfile, (char *) node);
         return (I_ERROR);
         }
      }
   if (rlse_blk(bfile, (char *) node) == I_ERROR)
      ERR2X(EC_RLSEBLK, CC_RESYNCPARENT)
   return (r_status);
   } /* add_children */
/**/
/*
        loc_start - locate start of re-sync for parent

        Loc_start locates the first child node that needs to be added to
        the parent and its position in the parent.

        Return OK or ERROR.
*/

static int loc_start(Cbtree *tree, Node *par_node, int par_pos,
                     int *start, Blk_Nbr *child)
   /* tree     = tree containing nodes */
   /* par_node = parent node buffer */
   /* par_pos  = position of key insertion */
   /* start    = insertion point of child */
   /* child    = first child to insert */
{
   Node *node;          /* child node buffer */
   Bfile *bfile;        /* buffered file pointer */

   bfile = (tree->ctl)->bfile;
   *start = par_pos - 1;
   if (*start < -1) *start = -1;
   if (*start == -1) *child = par_node->low_path;
   else *child = Ikey_Child(par_node, *start);
   if ((node = (Node *) get_blk(bfile, *child)) == (Node *) NULL)
      ERR2X(EC_GETBLK, CC_RESYNCPARENT)
   *child = node->right;
   (*start)++;
   if ((*start < par_node->key_cnt) && (*child == Ikey_Child(par_node, *start))) {
      if (rlse_blk(bfile, (char *) node) == I_ERROR)
         ERR2X(EC_RLSEBLK, CC_RESYNCPARENT)
      if ((node = (Node *) get_blk(bfile, *child)) == (Node *) NULL)
         ERR2X(EC_GETBLK, CC_RESYNCPARENT)
      *child = node->right;
      (*start)++;
      }
   if (rlse_blk(bfile, (char *) node) == I_ERROR)
      ERR2X(EC_RLSEBLK, CC_RESYNCPARENT)
   return (I_OK);
   } /* loc_start */
/**/
/*
        loc_stop - locate end of re-sync for parent

        Loc_stop locates the block number of the last child to be
        processed to re-sync the parent with its children.

        Return OK or ERROR.
*/

static int loc_stop(Cbtree *tree, Node *par_node, int start,
                    Blk_Nbr *sync_child)
   /* tree       = tree containing nodes */
   /* par_node   = parent node buffer */
   /* start      = insertion point of child */
   /* sync_child = last child to process */
{
   if (start < par_node->key_cnt) *sync_child = Ikey_Child(par_node, start);
   else {
      if (par_node->right == (Blk_Nbr) 0) *sync_child = (Blk_Nbr) 0;
      else {
         if (B_get_low_path(tree, par_node->right, sync_child) == I_ERROR) {
            SET_CC(CC_RESYNCPARENT)
            return (I_ERROR);
            }
         }
      }
   return (I_OK);
   } /* loc_stop */
/**/
/*
        B_add_low_key - add key from low path leaf

        The insertion position of the low path is returned in pos.
        Return OK, SPLIT, SHIFT, SHIFT_INS, or ERROR.
*/

int B_add_low_key(Cbtree *tree, Blk_Nbr blk_nbr, Node *node, Blk_Nbr child,
   Blk_Nbr parent, int par_pos, int *pos)
   /* tree     = tree containing nodes */
   /* blk_nbr  = block to add key to */
   /* node     = node buffer of blk_nbr */
   /* child    = child block to get low key from */
   /* parent   = parent of block to add to */
   /* par_pos  = position of blk_nbr in parent */
   /* pos      = OUTPUT insert pos of low key */
{
   Blk_Nbr low_path_blk;      /* block number of low path node */
   Node *low_node;            /* node buffer of low path node */
   Key_Info key_info;         /* key to insert into root */
   int status;                /* subroutine return status */
   Bfile *bfile;              /* buffered file pointer */

   CB_ASSERT(B_is_write_locked(tree->ctl));
   bfile = (tree->ctl)->bfile;
   if (B_loc_low_blk(tree, child, &low_path_blk) == I_ERROR)
      return (I_ERROR);
   if ((low_node = (Node *) get_blk(bfile, low_path_blk)) == (Node *) NULL)
      ERR2X(EC_GETBLK, CC_RESYNCPARENT)
   CB_ASSERT(low_node->key_cnt > 0);
   key_info.key = Lkey_Addr(low_node, 0);
   key_info.keylen = Lkey_Len(low_node, 0);
   key_info.item = Lkey_Item(low_node, 0);
   key_info.dup = I_YES;
   key_info.child = child;
   if ((status = B_ins_node(tree, &key_info, blk_nbr, parent, par_pos))
      == I_ERROR) {
      rlse_blk(bfile, (char *) low_node);
      return (I_ERROR);
      }
   if (rlse_blk(bfile, (char *) low_node) == I_ERROR)
      ERR2X(EC_RLSEBLK, CC_RESYNCPARENT)
   B_key_info_loc(tree, &key_info, node, pos);
   return (status);
   } /* B_add_low_key */

/* end cbinsert.c */
