/*      Copyright (c) 1994 Mix Software, Inc.

        cbkeycmp.c - CBT cbkeycmp function and key handling functions
*/
#include "btree.h"
#include "cbtree.h"
#include "cberr.h"
#define min(a,b)        ((a) < (b) ? (a) : (b))

/*
        Defined Functions (INTERNAL functions for use by CBT only)

        cbkeycmp                compare two CBT keys
        B_key_item_cmp          INTERNAL compare key and item
        B_key_info_loc          INTERNAL locate key from key_info in node
        B_is_dup_error          INTERNAL does key violate dup flag
        B_is_k_info_dup         INTERNAL is key_info key a duplicate
        B_node_keycmp           INTERNAL does key_info match key in node
        B_key_match             INTERNAL do two keys match
        B_cur_key_match         INTERNAL does key match current key
        B_cmp_low_path          INTERNAL compare key_info to low_path key
        B_loc_key               INTERNAL locate key in node
*/
#if ! defined(I_UNIX)
extern int memicmp (void *addr1, void *addr2, unsigned n);
#endif

/**/
/*
        cbkeycmp - compare two keys.

        cbkeycmp does a byte by byte compare of the two byte strings.  The
        compare is done according to the ASCII character set (for example,
        the space character ' ' is less than capital 'A').  If one key is a
        prefix of the other key then the shorter one is considered less than
        the longer one.

        This function is the default function used for key comparison if no
        function is specified to cbopen or cbcreate.

        Return: I_LESS    if keya <  keyb
                I_EQUAL   if keya == keyb
                I_GREATER if keya >  keyb
*/

I_EXPORT int I_ENTRY cbkeycmp(char *keya, int lena, char *keyb, int lenb)
{
   int status;         /* compare status */

   /* If one of the keys is zero length, return immediate answer. */

   if (lena == 0) return ((lenb == 0) ? I_EQUAL : I_LESS);
   else if (lenb == 0) return (I_GREATER);

   /* Pick shorter length and compare the first k bytes of the keys. */
   status = memicmp(keya, keyb, min (lena, lenb));
   if (status < 0) return (I_LESS);
   if (status > 0) return (I_GREATER);

   /* Strings are equal to this point, so resolve based on length. */
   return ((lena == lenb) ? I_EQUAL : ((lena < lenb) ? I_LESS : I_GREATER));
   } /* cbkeycmp */
/**/
/*
        B_key_item_cmp - compare key and item

        Return: I_LESS    if keya/itema <  keyb/itemb
                I_EQUAL   if keya/itema == keyb/itemb
                I_GREATER if keya/itema >  keyb/itemb
*/

int B_key_item_cmp(Cbtree *tree, char *keya, int lena, Blk_Nbr itema,
      char *keyb, int lenb, Blk_Nbr itemb)
{
   int status;         /* compare status */

   if ((status = (*(tree->keycmp))(keya, lena, keyb, lenb)) != I_EQUAL)
      return (status);
   if (itema < itemb) return (I_LESS);
   if (itema > itemb) return (I_GREATER);
   return (I_EQUAL);
   } /* B_key_item_cmp */
/**/
/*
        B_key_info_loc - locate key from key_info in node

        The location of the key is returned in pos.
        Return comparison status (I_LESS, I_EQUAL, or I_GREATER).
*/

int B_key_info_loc(Cbtree *tree, Key_Info *key_info, Node *node, int *pos)
{
   return (B_loc_key(tree, key_info->key, key_info->keylen, key_info->item,
           node, pos));
   } /* B_key_info_loc */
/**/
/*
        B_is_dup_error - is key in violation of dup flag within key_info

        Return I_YES or I_NO.
*/


int B_is_dup_error(Cbtree *tree, Key_Info *key_info, Node *node, int pos)
{
   Node    *nextnode;            /* node buffer */
   int     status;
   Blk_Nbr nextblock;
   Bfile *bfile;              /* buffered file pointer */

   CB_ASSERT(pos >= 0 && pos <= node->key_cnt);
   if (key_info->dup != I_NO) return (I_NO);
   if (pos < node->key_cnt) {
      if (B_node_keycmp(tree, key_info, node, pos) == I_YES) return (I_YES);
      }

   if (pos > 0 && pos <= node->key_cnt) {
      if (B_node_keycmp(tree, key_info, node, pos-1) == I_YES) return I_YES;
      if (pos == node->key_cnt) nextblock = node->right;
      else return (I_NO);
      }
   else if (pos == 0) nextblock = node->left;
   else return (I_NO);

   /* check the node to the left or right */
   if (nextblock == (Blk_Nbr) 0) return (I_NO);  /* none */
   CB_ASSERT(B_is_read_locked(tree->ctl));
   bfile = tree->ctl->bfile;
   if ((nextnode = (Node *) get_blk(bfile, nextblock)) == NULL)
      ERR1X(EC_GETBLK)
   if (pos == 0) pos = nextnode->key_cnt - 1; else pos = 0;
   CB_ASSERT(pos >= 0);
   if (pos < 0) status = I_NO;
   else status = B_node_keycmp(tree, key_info, nextnode, pos);
   rlse_blk(bfile, (char *) nextnode);
   return (status);
   } /* B_is_dup_error */
/**/
/*
        B_is_k_info_dup - is key_info key a duplicate

        B_is_k_info_dup looks at the keys at pos and pos - 1 to determine
        if it is a duplicate of the key_info key.  If it is, the offset
        of the duplicate key is returned.

        Return I_YES or I_NO.
*/

int B_is_k_info_dup(Cbtree *tree, Key_Info *key_info, Node *node, int pos,
      int *offset)
{
   CB_ASSERT(pos >= 0 && pos <= node->key_cnt);
   if (pos < node->key_cnt) {
      if (node->low_path == LEAF_NODE) *offset = Lkey_Offset(node, pos);
      else *offset = Ikey_Offset(node, pos);
      if (B_node_keycmp(tree, key_info, node, pos) == I_YES) return (I_YES);
      }
   if (pos > 0) {
      if (node->low_path == LEAF_NODE) *offset = Lkey_Offset(node, pos - 1);
      else *offset = Ikey_Offset(node, pos - 1);
      if (B_node_keycmp(tree, key_info, node, pos - 1) == I_YES) return (I_YES);
      }
   return (I_NO);
   } /* B_is_k_info_dup */
/**/
/*
        B_node_keycmp - does key_info match key in node at pos

        Return I_YES or I_NO.
*/

int B_node_keycmp(Cbtree *tree, Key_Info *key_info, Node *node, int pos)
{
   CB_ASSERT(pos >= 0 && pos < node->key_cnt);
   if (pos < 0 || pos >= node->key_cnt) return (I_NO);
   if (node->low_path == LEAF_NODE) {
      if ((*(tree->keycmp))(key_info->key, key_info->keylen,
         Lkey_Addr(node, pos), Lkey_Len(node, pos)) == I_EQUAL) return (I_YES);
      else return (I_NO);
      }
   else {
      if ((*(tree->keycmp))(key_info->key, key_info->keylen,
         Ikey_Addr(node, pos), Ikey_Len(node, pos)) == I_EQUAL) return (I_YES);
      else return (I_NO);
      }
   } /* B_node_keycmp */
/**/
/*
        B_key_match - are two keys equal

        Return I_YES or I_NO.
*/


int B_key_match(Cbtree *tree, Node *nodea, int posa, Node *nodeb, int posb)
{
   char *keya;                /* key a */
   int lena;                  /* length of key a */
   char *keyb;                /* key b */
   int lenb;                  /* length of key b */

   if (nodea->low_path == LEAF_NODE) {
      keya = Lkey_Addr(nodea, posa);
      lena = Lkey_Len(nodea, posa);
      }
   else {
      keya = Ikey_Addr(nodea, posa);
      lena = Ikey_Len(nodea, posa);
      }
   if (nodeb->low_path == LEAF_NODE) {
      keyb = Lkey_Addr(nodeb, posb);
      lenb = Lkey_Len(nodeb, posb);
      }
   else {
      keyb = Ikey_Addr(nodeb, posb);
      lenb = Ikey_Len(nodeb, posb);
      }
   if ((*(tree->keycmp))(keya, lena, keyb, lenb) == I_EQUAL) return (I_YES);
   else return (I_NO);
   } /* B_key_match */
/**/
/*
        B_cur_key_match - does key match current key

        Return I_YES or I_NO.
*/

int B_cur_key_match(Cbtree *tree, char *key, int keylen)
{
   Node *node;                /* node buffer */
   char *nkey;                /* key in node */
   int nkeylen;               /* length of key in node */
   int status;                /* key compare status */
   Bfile *bfile;              /* buffered file pointer */

   bfile = (tree->ctl)->bfile;
   if ((node = (Node *) get_blk(bfile, tree->cur_key.block)) == NULL)
      ERR2X(EC_GETBLK, CC_CURKEY)
   if ((tree->cur_key.index < 0) || (tree->cur_key.index >= node->key_cnt)) {
      SET_EC(EC_BADCURSOR)
      SET_CC(CC_CURKEY)
      rlse_blk(bfile, (char *) node);
      return (I_ERROR);
      }
   CB_ASSERT(node->low_path == LEAF_NODE);
   nkey = (char *) node + Lkey_Offset(node, tree->cur_key.index);
   nkeylen = Lkey_Len(node, tree->cur_key.index);
   if ((*(tree->keycmp))(key, keylen, nkey, nkeylen) == I_EQUAL) status = I_YES;
   else status = I_NO;
   rlse_blk(bfile, (char *) node);
   return (status);
   } /* B_cur_key_match */
/**/
/*
        B_cmp_low_path - compare key_info to low_path key

        B_cmp_low_path locates the key at the end of the low path and compares
        the key_info key to it.  The compare status is returned in status.

        Return I_OK or I_ERROR.
*/

int B_cmp_low_path(Cbtree *tree, Key_Info *key_info, Node *node, int *status)
{
   Blk_Nbr low_blk_nbr;          /* block at end of low path */
   Node *low_node;               /* node buffer of low path node */
   Bfile *bfile;                 /* buffered file pointer */

   bfile = (tree->ctl)->bfile;
   if (node->key_cnt == -1) {
      *status = I_LESS;
      return (I_OK);
      }
   if (B_loc_low_blk(tree, node->low_path, &low_blk_nbr) == I_ERROR)
      return (I_ERROR);
   if ((low_node = (Node *) get_blk(bfile, low_blk_nbr)) == (Node *) NULL)
      ERR2X(EC_GETBLK, CC_LOWBLOCK)
   *status = B_key_item_cmp(tree, key_info->key, key_info->keylen,
      key_info->item, Lkey_Addr(low_node, 0), Lkey_Len(low_node, 0),
      Lkey_Item(low_node, 0));
   if (rlse_blk(bfile, (char *) low_node) == I_ERROR)
      ERR2X(EC_RLSEBLK, CC_LOWBLOCK)
   return (I_OK);
   } /* B_cmp_low_path */
/**/
/*
        B_loc_key - locate key in node or leaf

        B_loc_key locates the key in the node or leaf, or the position where
        the key would go.  The position is returned in pos.

        Return comparison status (I_LESS, I_EQUAL, or I_GREATER).
*/

int B_loc_key(Cbtree *tree, char *key, int keylen, Item item, Node *node,
   int *pos)
{
   int lower;                  /* lower bound on search */
   int upper;                  /* upper bound on search */
   int status = I_LESS;        /* key compare status */
   L_Pairs *cl;                /* current key in leaf node */
   I_Pairs *ci;                /* current key in interior node */
   int key_index = 0;          /* current key index */

   *pos = 0;
   if (node->key_cnt <= 0) return (I_GREATER); /* empty node */
   lower = 0;
   upper = node->key_cnt - 1;
   if (node->low_path == LEAF_NODE) {
      while (upper >= lower) {   /* binary search for match */
         key_index = (upper + lower) / 2;
         cl = &(node->pairs.l_node[key_index]);
         status = B_key_item_cmp(tree, key, keylen, item,
            (char *) node + cl->key_offset, cl->key_len, cl->key_item);
         switch (status) {
            case I_LESS:
               upper = key_index - 1;
               break;
            case I_GREATER:
               lower = key_index + 1;
               break;
            case I_EQUAL:
               *pos = key_index;
               return (I_EQUAL);
            }
         }
      }
   else {
      while (upper >= lower) {
         key_index = (upper + lower) / 2;
         ci = &(node->pairs.i_node[key_index]);
         status = B_key_item_cmp(tree, key, keylen, item,
            (char *) node + ci->key_offset, ci->key_len, ci->key_item);
         switch (status) {
            case I_LESS:
               upper = key_index - 1;
               break;
            case I_GREATER:
               lower = key_index + 1;
               break;
            case I_EQUAL:
               *pos = key_index;
               return (I_EQUAL);
            }
         }
      }

   /* key not found, return its insertion position */
   if (status == I_GREATER) key_index++;
   if (key_index < node->key_cnt) {
      if (node->low_path == LEAF_NODE) {
         cl = &(node->pairs.l_node[key_index]);
         status = B_key_item_cmp(tree, key, keylen, item,
            (char *) node + cl->key_offset, cl->key_len, cl->key_item);
         }
      else {
         ci = &(node->pairs.i_node[key_index]);
         status = B_key_item_cmp(tree, key, keylen, item,
            (char *) node + ci->key_offset, ci->key_len, ci->key_item);
         }
      }
   *pos = key_index;
   return (status);
   } /* B_loc_key */

/* end cbkeycmp.c */
