/*      Copyright (c) 1995 Mix Software, Inc.

        os2io.c - module to read and write blocks from an open file
                (internal use only)

        The file is assumed to be a collection of blksize-sized blocks.
        The buffer must be large enough to hold blksize bytes.
*/
#define INCL_DOSFILEMGR
#define INCL_DOSPROCESS
#include <os2.h>
#include <string.h>
#include "cdtdef.h"
#include "blockio.h"

#if defined(I_DLL)
#define USE_STD_IO    0
#else
#define USE_STD_IO    1
#endif

#if USE_STD_IO
#include <stdio.h>
#else
#define STDOUT_HANDLE 1
#endif

int _cbt_doserrno = 0;
int _b_no_flush = 0;

/*
        Defined functions (CBT internal use only)

        cb_openfile               open a random file
        cb_close                  close a file
        cb_read                   read a random area from a file
        cb_readblk                read block from file
        cb_readseq                read sequentially
        cb_write                  write to a file
        cb_writeblk               write block to file
        cb_writeseq               write sequentially
        cb_commitblk              commit updates to disk
        cb_lock_area              lock a portion of a file
        cb_unlock_area            unlock a portion of a file
        cb_delay                  delay (before lock retries)
        cb_prints                 print a string to stdout
        cb_can_share              can files be shared?
*/
int cb_read(int fd, unsigned long location, int size, void *buffer);
int cb_write(int fd, unsigned long location, int size, void *buffer);

/**/
/* ------------------------------------------------------------------------
        cb_openfile - open a block oriented file

        Filename is the name of the file to be opened.  Share is either I_YES
        or I_NO, (1 or 0).  If share is I_YES, the file is opened for shared
        access so that multiple programs can access the file at the same time.

        Return a non-negative integer on success or -1 on error.  In the
        event of an error, errno is also set.
*/

int cb_openfile(char *filename, int mode, int share)
{
   APIRET rc;
   HFILE fd;
   ULONG action;
   ULONG attrib;
   ULONG modeflg;

   if (mode == CBT_OPEN_RDONLY) attrib = FILE_READONLY;
   else attrib = FILE_NORMAL;
   modeflg = OPEN_ACCESS_READWRITE | OPEN_FLAGS_NOINHERIT | OPEN_FLAGS_RANDOM;
   if (share) modeflg |= OPEN_SHARE_DENYNONE;
   else modeflg |= OPEN_SHARE_DENYREADWRITE;

   rc = DosOpen(filename, &fd, &action, 0, attrib,
      OPEN_ACTION_OPEN_IF_EXISTS | OPEN_ACTION_FAIL_IF_NEW,
      modeflg, NULL);

   if (rc != 0) {
      _cbt_doserrno = rc;
      return -1;
      }
   return fd;
} /* cbopenfile */
/**/
/* ------------------------------------------------------------------------
        cb_create - create a file and open it for exclusive write

        Filename is the name of the file to be opened.
*/

int cb_create(char *filename)
{
   APIRET rc;
   HFILE fd;
   ULONG action;

   rc = DosOpen(filename, &fd, &action, 0, FILE_NORMAL,
      OPEN_ACTION_REPLACE_IF_EXISTS | OPEN_ACTION_CREATE_IF_NEW,
      OPEN_ACCESS_READWRITE | OPEN_FLAGS_NOINHERIT | OPEN_FLAGS_RANDOM
       | OPEN_SHARE_DENYREADWRITE, NULL);

   if (rc != 0) {
      _cbt_doserrno = rc;
      return -1;
      }
   return fd;
} /* cb_create */
/**/
/* ------------------------------------------------------------------------

        cbcloseblk - close a block oriented file

        Return 0 on success or -1 on error.
*/

int cb_closeblk(int fd)
{
   APIRET rc;
   rc = DosClose(fd);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return -1;
      }
   return 0;
} /* cb_closeblk */
/**/
/* ------------------------------------------------------------------------

        cbreadblk - read one block of blksize from file into buffer

        Return OK or ERROR.
*/
int cb_readblk(int fd, Blk_Nbr blocknbr, int blksize, void *buffer)
{
   return cb_read(fd, blocknbr * blksize, blksize, buffer);
   }

/* ------------------------------------------------------------------------

        cb_read  - random read

        Return OK or ERROR.
*/

int cb_read(int fd, unsigned long location, int size, void *buffer)
{
   APIRET rc;
   ULONG  readlength;
   ULONG  newlocation;

   /* seek to position */
   rc = DosSetFilePtr(fd, location, FILE_BEGIN, &newlocation);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return I_ERROR;
      }

   /* read data from file */
   rc = DosRead(fd, buffer, size, &readlength);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return I_ERROR;
      }
   if (readlength < (unsigned) size) {
      _cbt_doserrno = EC_DOS_READERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cbreadblk */

/* ------------------------------------------------------------------------

        cb_readseq  - sequential read

        Return OK or ERROR.
*/

int cb_readseq(int fd, int size, void *buffer)
{
   APIRET rc;
   ULONG  readlength;

   rc = DosRead(fd, buffer, size, &readlength);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return I_ERROR;
      }
   if (readlength < (unsigned) size) {
      _cbt_doserrno = EC_DOS_READERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cbreadseq */
/**/
/* ------------------------------------------------------------------------

        cbwriteblk - write one block of blksize to file from buffer

        Return OK or ERROR.
*/

int cb_writeblk(int fd, Blk_Nbr blocknbr, int blksize, void *buffer)
{
   return cb_write(fd, blocknbr * blksize, blksize, buffer);
   }

/* ------------------------------------------------------------------------

        cb_write - random write

        Return OK or ERROR.
*/

int cb_write(int fd, unsigned long location, int size, void *buffer)
{
   APIRET rc;
   ULONG  newlocation;
   ULONG  count;

   /* seek to position */
   rc = DosSetFilePtr(fd, location, FILE_BEGIN, &newlocation);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return I_ERROR;
      }

   /* write the data */
   rc = DosWrite(fd, buffer, size, &count);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return I_ERROR;
      }
   if (count < (unsigned) size) {
      _cbt_doserrno = EC_DOS_WRITERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cb_write */

/* ------------------------------------------------------------------------

        cb_writeseq - sequential write

        Return OK or ERROR.
*/

int cb_writeseq(int fd, int size, void *buffer)
{
   APIRET rc;
   ULONG  count;
   rc = DosWrite(fd, buffer, size, &count);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return I_ERROR;
      }
   if (count < (unsigned) size) {
      _cbt_doserrno = EC_DOS_WRITERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cb_writeseq */

/**/
/* ------------------------------------------------------------------------

        cb_remove - delete a file

        Return OK or ERROR.
*/

int cb_remove(char *filename)
{
   APIRET rc;

   rc = DosDelete(filename);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return I_ERROR;
      }
   return I_OK;
   } /* cb_remove */

/**/
/* ------------------------------------------------------------------------

        cb_lock_area - put a record lock on a file region

        Return OK or ERROR.
*/

int cb_lock_area(int fd, unsigned long location, long size, int tries,
   long delay, long timeout)
{
   APIRET rc;
   FILELOCK area;

   if (tries * delay < timeout) timeout = tries * delay;
   area.lOffset = location;
   area.lRange = size;

   rc = DosSetFileLocks(fd, NULL, &area, timeout, 0);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return I_ERROR;
      }
   return I_OK;
} /* cb_lock_area */
/**/
/* ------------------------------------------------------------------------

        cb_unlock_area - remove a record lock from a file region

        Return OK or ERROR.
*/

int cb_unlock_area(int fd, unsigned long location, long size)
{
   APIRET rc;
   FILELOCK area;

   area.lOffset = location;
   area.lRange = size;

   rc = DosSetFileLocks(fd, &area, NULL, 0, 0);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return I_ERROR;
      }
   return I_OK;
} /* cb_unlock_area */
/**/
/* ------------------------------------------------------------------------

        cb_commitblk -   flush any system buffer associated with specified
                         file descriptor

        Return OK or ERROR.
*/

int cb_commitblk(int fd)
{
   APIRET rc;

   if (_b_no_flush) return I_OK;
   rc = DosResetBuffer(fd);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return I_ERROR;
      }
   return I_OK;
   }
/**/
/* ------------------------------------------------------------------------

        cb_delay - time delay before lock retry

        always returns OK
*/

int cb_delay(long time)
{
   DosSleep(time);
   return I_OK;
} /* cb_delay */

/**/
/* ------------------------------------------------------------------------

        cb_prints - print a string to stdout

*/
#if USE_STD_IO
int I_ENTRY cb_prints(const char *string)
{
    if (fputs(string, stdout) != 0)
        return I_ERROR;

   return I_OK;
} /* cb_prints */

#else
int I_ENTRY cb_prints(const char *string)
{
   char *ptr;
   char crlf[3] = "\r\n";
   int  len;

   if (string == NULL) return I_ERROR;
   ptr = strchr(string, '\n');
   if (ptr == NULL)        /* no newlines to translate */
      return cb_writeseq(STDOUT_HANDLE, strlen(string), (void *)string);
   else {
      while (ptr != NULL) {
         if (ptr != string) {
            if (cb_writeseq(STDOUT_HANDLE, ptr - string, (void *)string) != I_OK)
               return I_ERROR;
            string = ptr;
            }
         if (cb_writeseq(STDOUT_HANDLE, 2, (void *)crlf) != I_OK) return I_ERROR;
         string++;
         ptr = strchr(string, '\n');
         }
      if ((len = strlen(string)) > 0) {
         if (cb_writeseq(STDOUT_HANDLE, len, (void *)string) != I_OK)
            return I_ERROR;
         }
      }
   return I_OK;
} /* cb_prints */
#endif
/**/

/* ------------------------------------------------------------------------

        cb_can_share - can files be shared?

        Return I_YES or I_NO.
*/

int cb_can_share(void)
{
   return I_YES;
} /* cb_can_share */
/**/

/* ------------------------------------------------------------------------

        cb_filesize - determine the size of an open file

        Return OK or ERROR.
*/

int cb_filesize(int fd, long *size)
{
   APIRET rc;

   /* seek to position */
   rc = DosSetFilePtr(fd, 0, FILE_END, (PULONG) size);
   if (rc != 0) {
      _cbt_doserrno = rc;
      return I_ERROR;
      }
   return (I_OK);
   } /* cb_filesize */

/* ------------------------------------------------------------------------

        cbset_print - set the print function
        Return OK or ERROR.
*/

I_EXPORT Puts_Fn_Ptr I_ENTRY cbset_print(Puts_Fn_Ptr printfn)
{
   Puts_Fn_Ptr old;
   old = _cb_prints;
   _cb_prints = printfn;
   return old;
   } /* cbset_print */

/* end os2io.c */
