/*      Copyright (c) 1996 Mix Software, Inc.

        uio.c - module to read and write blocks from an open file
                (internal use only)

        The file is assumed to be a collection of blksize-sized blocks.
        The buffer must be large enough to hold blksize bytes.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cdtdef.h"
#include "blockio.h"

int _cbt_doserrno = 0;
int _b_no_flush = 0;

/*
        Defined functions (CBT internal use only)

        cb_openfile               open a random file
        cb_close                  close a file
        cb_read                   read a random area from a file
        cb_readblk                read block from file
        cb_readseq                read sequentially
        cb_write                  write to a file
        cb_writeblk               write block to file
        cb_writeseq               write sequentially
        cb_commitblk              commit updates to disk
        cb_lock_area              lock a portion of a file
        cb_unlock_area            unlock a portion of a file
        cb_delay                  delay (before lock retries)
        cb_prints                 print a string to stdout
        cb_can_share              can files be shared?
*/
int cb_read(int fd, unsigned long location, int size, void *buffer);
int cb_write(int fd, unsigned long location, int size, void *buffer);

/**/
/* ------------------------------------------------------------------------
        cb_openfile - open a block oriented file

        Filename is the name of the file to be opened.  Share is either I_YES
        or I_NO, (1 or 0).  If share is I_YES, the file is opened for shared
        access so that multiple programs can access the file at the same time.

        Return a non-negative integer on success or -1 on error.  In the
        event of an error, errno is also set.
*/

int cb_openfile(char *filename, int mode, int share)
{
   int   fd;
   int   flags;
   struct flock lock;

   if (mode == CBT_OPEN_RDONLY) flags = O_RDONLY;
   else flags = O_RDWR;
   fd = open(filename, flags);
   if (fd < 0) {
      _cbt_doserrno = errno;
      return -1;
      }

   /* enforce exclusive access:  When a file is opened unshared, the  */
   /* entire file is write locked.  A shared file has the first byte  */
   /* read locked.  This prevents any other process from opening it   */
   /* as unshared.  */
   lock.l_whence = SEEK_SET;
   lock.l_start = 0;
   if (share) {
      lock.l_type = F_RDLCK;
      lock.l_len = 1;
      }
   else {
      /* fcntl does not allow a read only file to set a write lock */
      lock.l_type = (mode == CBT_OPEN_RDONLY) ? F_RDLCK : F_WRLCK;
      lock.l_len = 0;
      }
   if (fcntl(fd, F_SETLK, &lock) < 0) {
      _cbt_doserrno = errno;
      close(fd);
      return -1;
      }
   return fd;
} /* cbopenfile */
/**/
/* ------------------------------------------------------------------------
        cb_create - create a file and open it for exclusive write

        Filename is the name of the file to be opened.
*/

int cb_create(char *filename)
{
   int    fd;
   struct flock lock;

   fd = open(filename, O_RDWR | O_CREAT | O_TRUNC);
   if (fd < 0) {
      _cbt_doserrno = errno;
      return -1;
      }

   /* set manditory file/record locking */
   fchmod(fd, S_ISGID | S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
   /* obtain exclusive access */
   lock.l_whence = SEEK_SET;
   lock.l_start = 0;
   lock.l_type = F_WRLCK;
   lock.l_len = 0;
   if (fcntl(fd, F_SETLK, &lock) < 0) {
      _cbt_doserrno = errno;
      close(fd);
      return -1;
      }
   return fd;
} /* cb_create */
/**/
/* ------------------------------------------------------------------------

        cbcloseblk - close a block oriented file

        Return 0 on success or -1 on error.
*/

int cb_closeblk(int fd)
{
   if (close(fd) < 0) {
      _cbt_doserrno = errno;
      return -1;
      }
   return 0;
} /* cb_closeblk */
/**/
/* ------------------------------------------------------------------------

        cbreadblk - read one block of blksize from file into buffer

        Return OK or ERROR.
*/
int cb_readblk(int fd, Blk_Nbr blocknbr, int blksize, void *buffer)
{
   return cb_read(fd, blocknbr * blksize, blksize, buffer);
   }

/* ------------------------------------------------------------------------

        cb_read  - random read

        Return OK or ERROR.
*/

int cb_read(int fd, unsigned long location, int size, void *buffer)
{
   long length;

   /* seek to position */
   if (lseek(fd, location, SEEK_SET) == -1) {
      _cbt_doserrno = errno;
      return I_ERROR;
      }

   /* read data from file */
   length = read(fd, buffer, size);
   if (length < 0) {
      _cbt_doserrno = errno;
      return I_ERROR;
      }
   if (length < size) {
      _cbt_doserrno = EC_DOS_READERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cbreadblk */

/* ------------------------------------------------------------------------

        cb_readseq  - sequential read

        Return OK or ERROR.
*/

int cb_readseq(int fd, int size, void *buffer)
{
   long length;

   length = read(fd, buffer, size);
   if (length < 0) {
      _cbt_doserrno = errno;
      return I_ERROR;
      }
   if (length < size) {
      _cbt_doserrno = EC_DOS_READERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cbreadseq */
/**/
/* ------------------------------------------------------------------------

        cbwriteblk - write one block of blksize to file from buffer

        Return OK or ERROR.
*/

int cb_writeblk(int fd, Blk_Nbr blocknbr, int blksize, void *buffer)
{
   return cb_write(fd, blocknbr * blksize, blksize, buffer);
   }

/* ------------------------------------------------------------------------

        cb_write - random write

        Return OK or ERROR.
*/

int cb_write(int fd, unsigned long location, int size, void *buffer)
{
   long   count;

   /* seek to position */
   if (lseek(fd, location, SEEK_SET) == -1) {
      _cbt_doserrno = errno;
      return I_ERROR;
      }

   /* write data to file */
   count = write(fd, buffer, size);
   if (count < 0) {
      _cbt_doserrno = errno;
      return I_ERROR;
      }
   if (count < size) {
      _cbt_doserrno = EC_DOS_WRITERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cb_write */

/* ------------------------------------------------------------------------

        cb_writeseq - sequential write

        Return OK or ERROR.
*/

int cb_writeseq(int fd, int size, void *buffer)
{
   long   count;
   count = write(fd, buffer, size);
   if (count < 0) {
      _cbt_doserrno = errno;
      return I_ERROR;
      }
   if (count < size) {
      _cbt_doserrno = EC_DOS_WRITERR;
      return I_ERROR;
      }
   return I_OK;
   } /* cb_writeseq */

/**/
/* ------------------------------------------------------------------------

        cb_remove - delete a file

        Return OK or ERROR.
*/

int cb_remove(char *filename)
{
   if (unlink(filename) < 0) {
      _cbt_doserrno = errno;
      return I_ERROR;
      }
   return I_OK;
   } /* cb_remove */

/**/
/* ------------------------------------------------------------------------

        cb_lock_area - put a record lock on a file region

        Return OK or ERROR.
*/

int cb_lock_area(int fd, unsigned long location, long size, int tries,
   long delay, long timeout)
{
   struct flock lock;
   struct timeval time, stoptime;
   struct timezone tz;
   int    locktype;
   int    usetimeout = 1;

   if (tries * delay < timeout) timeout = tries * delay;
   if ((fcntl(fd, F_GETFL) & O_ACCMODE) == O_RDONLY) locktype = F_RDLCK;
   else locktype = F_WRLCK;

   lock.l_whence = SEEK_SET;
   lock.l_start = location;
   lock.l_type = locktype;
   lock.l_len = size;
   if (fcntl(fd, F_SETLK, &lock) == 0) return I_OK;
   if (tries < 2) return I_ERROR;

   if (gettimeofday(&time, &tz) == -1) usetimeout = 0;
   else {
      stoptime.tv_sec = time.tv_sec + timeout / 1000;
      stoptime.tv_usec = time.tv_usec + 1000 * (timeout % 1000);
      while (stoptime.tv_usec > 1000000) {
         stoptime.tv_sec++;
         stoptime.tv_usec -= 1000000;
         }
      }

   while (--tries > 0) {
      usleep(delay*1000);
      lock.l_whence = SEEK_SET;
      lock.l_start = location;
      lock.l_type = locktype;;
      lock.l_len = size;
      if (fcntl(fd, F_SETLK, &lock) == 0) return I_OK;
      if (usetimeout && gettimeofday(&time, &tz) == 0) {
         if (time.tv_sec > stoptime.tv_sec) return I_ERROR;
         if (time.tv_sec == stoptime.tv_sec && time.tv_usec > stoptime.tv_usec)
            return I_ERROR;
         }
      }
   return I_OK;
} /* cb_lock_area */
/**/
/* ------------------------------------------------------------------------

        cb_unlock_area - remove a record lock from a file region

        Return OK or ERROR.
*/

int cb_unlock_area(int fd, unsigned long location, long size)
{
   struct flock lock;

   lock.l_whence = SEEK_SET;
   lock.l_start = location;
   lock.l_type = F_UNLCK;
   lock.l_len = size;
   if (fcntl(fd, F_SETLK, &lock) == 0) return I_OK;

   return I_ERROR;
} /* cb_unlock_area */
/**/
/* ------------------------------------------------------------------------

        cb_commitblk -   flush any system buffer associated with specified
                         file descriptor

        Return OK or ERROR.
*/

int cb_commitblk(int fd)
{
   return I_OK;
   }
/**/
/* ------------------------------------------------------------------------

        cb_delay - time delay before lock retry

        always returns OK
*/

int cb_delay(long time)
{
   usleep(time*1000);
   return I_OK;
} /* cb_delay */

/**/
/* ------------------------------------------------------------------------

        cb_prints - print a string to stdout

*/
int I_ENTRY cb_prints(const char *string)
{
    if (fputs(string, stdout) != 0)
        return I_ERROR;

   return I_OK;
} /* cb_prints */

/**/

/* ------------------------------------------------------------------------

        cb_can_share - can files be shared?

        Return I_YES or I_NO.
*/

int cb_can_share(void)
{
   return I_YES;
} /* cb_can_share */
/**/

/* ------------------------------------------------------------------------

        cb_filesize - determine the size of an open file

        Return OK or ERROR.
*/

int cb_filesize(int fd, long *size)
{
   long current;

   if ((current = lseek(fd, 0, SEEK_CUR)) < 0) return (I_ERROR);
   *size = lseek(fd, 0, SEEK_END);
   if (lseek(fd, current, SEEK_SET) < 0) return (I_ERROR);
   if (*size < 0) return I_ERROR;
   return (I_OK);
   } /* cb_filesize */

/* ------------------------------------------------------------------------

        cbset_print - set the print function
        Return OK or ERROR.
*/

I_EXPORT Puts_Fn_Ptr I_ENTRY cbset_print(Puts_Fn_Ptr printfn)
{
   Puts_Fn_Ptr old;
   old = _cb_prints;
   _cb_prints = printfn;
   return old;
   } /* cbset_print */

/* end os2io.c */
